unit HttpHelperU;

interface

uses
  idHTTP;

{$M+}

const
  HTTP_STATUS_OK = 200;
  HTTP_STATUS_CREATED = 201;
  HTTP_STATUS_UNAUTHORIZED = 401;
  HTTP_STATUS_FORBIDDEN = 403;
  HTTP_STATUS_NOT_FOUND = 404;
  HTTP_STATUS_METHOD_NOT_ALLOWED = 405;

type
  THttpHelper = class
  strict private
  type
    TRestError = record
    public
      LastHttpResponseCode: Integer;
      LastResponseText: string;
      LastException: string;
      procedure Clear;
      function ResponseOK: Boolean;
    end;

  var
    InternalHTTP: TIdHTTP;
    FRestError: TRestError;
    procedure UpdateLastError;
  public
    function HTTPPost(const aUrl: string): Boolean;
    function HTTPPut(const aUrl: string): Boolean;
    function HTTPGet(const aUrl: string): string;
    function HTTPDelete(const aUrl: string): Boolean;
    procedure ReCreateHTTP;
    constructor Create;
  published
    property RestError: TRestError read FRestError;
  end;

var
  HttpHelper: THttpHelper;

implementation

uses
  SysUtils, Classes;
{ THttpHelper }

constructor THttpHelper.Create;
begin
  inherited;
  ReCreateHTTP;
end;

function THttpHelper.HTTPDelete(const aUrl: string): Boolean;
begin
  FRestError.Clear;
  try
    try
      InternalHTTP.Delete(aUrl);
      Result := True;
    except
      on e: Exception do
      begin
        Result := False;
        FRestError.LastException := e.Message;
      end;
    end;
  finally
    UpdateLastError;
  end;
end;

function THttpHelper.HTTPGet(const aUrl: string): string;
begin
  FRestError.Clear;
  try
    try
      Result := InternalHTTP.Get(aUrl);
    except
      on e: Exception do
      begin
        Result := '';
        FRestError.LastException := e.Message;
      end;
    end;
  finally
    UpdateLastError;
  end;
end;

function THttpHelper.HTTPPost(const aUrl: string): Boolean;
var
  Dummy: TStringList;
begin
  Dummy := TStringList.Create;
  try
    try
      FRestError.Clear;
      InternalHTTP.Post(aUrl, Dummy);
      Result := True;
    except
      on e: Exception do
      begin
        FRestError.LastException := e.Message;
        Result := False;
      end;
    end;
  finally
    UpdateLastError;
    FreeAndNil(Dummy);
  end;
end;

function THttpHelper.HTTPPut(const aUrl: string): Boolean;
var
  Dummy: TMemoryStream;
begin
  Dummy := TMemoryStream.Create;
  try
    try
      InternalHTTP.Put(aUrl, Dummy);

      Result := True;
    except
      on e: Exception do
      begin
        FRestError.LastException := e.Message;
        Result := False;
      end;
    end;
  finally
    UpdateLastError;
    FreeAndNil(Dummy);
  end;
end;

procedure THttpHelper.ReCreateHTTP;
begin
  if Assigned(InternalHTTP) then
    FreeAndNil(InternalHTTP);
  InternalHTTP := TIdHTTP.Create(nil);
end;

procedure THttpHelper.UpdateLastError;
begin
  FRestError.LastHttpResponseCode := InternalHTTP.ResponseCode;
  FRestError.LastResponseText := InternalHTTP.ResponseText;
end;

{ THttpHelper.TRestError }

procedure THttpHelper.TRestError.Clear;
begin
  LastHttpResponseCode := -1;
  LastResponseText := '';
  LastException := '';
end;

function THttpHelper.TRestError.ResponseOK: Boolean;
begin
  Result := LastHttpResponseCode = HTTP_STATUS_OK;
end;

initialization

HttpHelper := THttpHelper.Create;

finalization

FreeAndNil(HttpHelper);

end.
