unit RESTWebModuleU;

interface

uses
  SysUtils, Classes, HTTPApp, RegularExpressions, Generics.Collections;

const
  HTTP_STATUS_OK = 200;
  HTTP_STATUS_CREATED = 201;
  HTTP_STATUS_UNAUTHORIZED = 401;
  HTTP_STATUS_FORBIDDEN = 403;
  HTTP_STATUS_NOT_FOUND = 404;
  HTTP_STATUS_METHOD_NOT_ALLOWED = 405;
  CONTENT_TYPE_TEXT = 'text/plain';
  CONTENT_TYPE_HTML = 'text/html';
  CONTENT_TYPE_JSON = 'application/json';

type
  TRESTRequestHandlerClass = class of TRESTRequestHandler; // metaclass type

  TRESTRequestHandler = class abstract
  private
    FPattern: string;
    FRegex: TRegEx;
  protected
    function GetPath: string; virtual; abstract;
    function HandleGet(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean; virtual;
    function HandlePost(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean; virtual;
    function HandleDelete(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean; virtual;
    function HandlePut(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean; virtual;
    function IsAuthenticated(const AuthenticationID: string; Response: TWebResponse): Boolean; virtual;
  public
    constructor Create;
    function Handle(Request: TWebRequest; Response: TWebResponse): Boolean;
    property Path: string read GetPath;
  end;

  TRESTWebModule = class(TWebModule)
    procedure RESTWebModuleDefaultHandlerAction(Sender: TObject; Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
  private
    class var FHandlers: TObjectList<TRESTRequestHandler>;
  public
    class procedure RegisterHandler(RequestHandlerClass: TRESTRequestHandlerClass);
  end;

var
  WebModuleClass: TComponentClass = TRESTWebModule;

implementation

uses
  Activex, StrUtils, ModelU;

{$R *.dfm}

class procedure TRESTWebModule.RegisterHandler(RequestHandlerClass: TRESTRequestHandlerClass);
begin
  if Assigned(TRESTWebModule.FHandlers) then
    TRESTWebModule.FHandlers.Add(RequestHandlerClass.Create);
end;

procedure TRESTWebModule.RESTWebModuleDefaultHandlerAction(Sender: TObject; Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
var
  RESTRequestHandler: TRESTRequestHandler;
begin
  for RESTRequestHandler in TRESTWebModule.FHandlers do
    if RESTRequestHandler.Handle(Request, Response) then
      exit;

  Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
  Response.ContentType := CONTENT_TYPE_TEXT;
  Response.Content := 'Requested uri is not registered with any handler';
  Handled := True;
end;

constructor TRESTRequestHandler.Create;
var
  Pattern: string;
begin
  inherited;
  Pattern := TRegEx.Escape(Path);
  Pattern := StringReplace(Pattern, '\*\*', '.*', [rfReplaceAll]);
  Pattern := StringReplace(Pattern, '\*', '[^/]*', [rfReplaceAll]);
  Pattern := StringReplace(Pattern, '\?', '.', [rfReplaceAll]);
  FPattern := Pattern;
  FRegex := TRegEx.Create('^' + Pattern + '$', [roIgnoreCase]);
end;

function TRESTRequestHandler.Handle(Request: TWebRequest; Response: TWebResponse): Boolean;
var
  Variables: TList<string>;

  // Adjacent * or ? values will be concatenated
  procedure GetVariables(const s: string);
  var
    i, j, p: integer;
  begin
    i := PosEx('*', s);
    if i < 1 then
      i := PosEx('?', s);

    if i > 0 then
    begin
      j := i;
      while (j <= Length(s)) and CharInSet(s[j], ['*', '?']) do
        Inc(j);

      if j <= Length(s) then
        p := PosEx(s[j], string(Request.PathInfo), i)
      else
        p := Length(Request.PathInfo) + 1;

      Variables.Add(Copy(string(Request.PathInfo), i, p - i));
      GetVariables(Copy(s, 1, i - 1) + Copy(string(Request.PathInfo), i, p - i) + Copy(s, j, Length(s)));
    end;
  end;

begin
  if not FRegex.IsMatch(string(Request.PathInfo)) then
    exit(False);

  Variables := TList<string>.Create;
  try
    CoInitialize(nil);
    GetVariables(Path);

    if Request.Method = 'GET' then
    begin
      if not HandleGet(Request, Response, Variables) then
        Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
    end
    else if Request.Method = 'POST' then
    begin
      if not HandlePost(Request, Response, Variables) then
        Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
    end
    else if Request.Method = 'DELETE' then
    begin
      if not HandleDelete(Request, Response, Variables) then
        Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
    end
    else if Request.Method = 'PUT' then
    begin
      if not HandlePut(Request, Response, Variables) then
        Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
    end
    else
      Response.StatusCode := HTTP_STATUS_METHOD_NOT_ALLOWED;
  finally
    Variables.Free;
    CoUninitialize;
  end;

  Result := True;
end;

function TRESTRequestHandler.HandleDelete(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean;
begin
  Result := False;
end;

function TRESTRequestHandler.HandleGet(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean;
begin
  Result := False;
end;

function TRESTRequestHandler.HandlePost(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean;
begin
  Result := False;
end;

function TRESTRequestHandler.HandlePut(Request: TWebRequest; Response: TWebResponse; Variables: TList<string>): Boolean;
begin
  Result := False;
end;

function TRESTRequestHandler.IsAuthenticated(const AuthenticationID: string; Response: TWebResponse): Boolean;
begin
  Result := Model.AdminCheckAuth(AuthenticationID);
  if not Result then
  begin
    Response.StatusCode := HTTP_STATUS_UNAUTHORIZED;
    Response.Content := 'Not a valid authorization identifier';
    Response.ContentType := CONTENT_TYPE_TEXT;
  end
  else
    Response.ContentType := CONTENT_TYPE_JSON;
end;

initialization

TRESTWebModule.FHandlers := TObjectList<TRESTRequestHandler>.Create(True);

finalization

if Assigned(TRESTWebModule.FHandlers) then
  FreeAndNil(TRESTWebModule.FHandlers);

end.
